import numpy as np
from scipy import sparse

#############################################
#      Generate random cone problems        #
#############################################


def gen_feasible(K, n, density):
    m = get_scs_cone_dims(K)
    z = np.random.randn(m)
    y = proj_dual_cone(z, K)  # y = s - z;
    s = y - z  # s = proj_cone(z,K)

    A = sparse.rand(m, n, density, format="csc")
    A.data = np.random.randn(A.nnz)
    x = np.random.randn(n)
    c = -np.transpose(A).dot(y)
    b = A.dot(x) + s

    data = {"A": A, "b": b, "c": c}
    return data, np.dot(c, x)


def gen_infeasible(K, n):
    m = get_scs_cone_dims(K)

    z = np.random.randn(m)
    y = proj_dual_cone(z, K)  # y = s - z;
    A = np.random.randn(m, n)
    A = (
        A - np.outer(y, np.transpose(A).dot(y)) / np.linalg.norm(y) ** 2
    )  # dense...

    b = np.random.randn(m)
    b = -b / np.dot(b, y)

    data = {"A": sparse.csc_matrix(A), "b": b, "c": np.random.randn(n)}
    return data


def gen_unbounded(K, n):
    m = get_scs_cone_dims(K)

    z = np.random.randn(m)
    s = proj_cone(z, K)
    A = np.random.randn(m, n)
    x = np.random.randn(n)
    A = A - np.outer(s + A.dot(x), x) / np.linalg.norm(x) ** 2
    # dense...
    c = np.random.randn(n)
    c = -c / np.dot(c, x)

    data = {"A": sparse.csc_matrix(A), "b": np.random.randn(m), "c": c}
    return data


def pos(x):
    return (x + abs(x)) / 2


def get_scs_cone_dims(K):
    l = K["z"] + K["l"]
    for i in range(0, len(K["q"])):
        l = l + K["q"][i]

    for i in range(0, len(K["s"])):
        l = l + get_sd_cone_size(K["s"][i])

    l = l + K["ep"] * 3
    l = l + K["ed"] * 3
    l = l + len(K["p"]) * 3
    return int(l)


def proj_dual_cone(z, c):
    return z + proj_cone(-z, c)


def get_sd_cone_size(n):
    return int((n * (n + 1)) / 2)


def proj_cone(z, c):
    z = np.copy(z)
    free_len = c["z"]
    lp_len = c["l"]
    q = c["q"]
    s = c["s"]
    p = c["p"]
    # free/zero cone
    z[0:free_len] = 0
    # lp cone
    z[free_len : lp_len + free_len] = pos(z[free_len : lp_len + free_len])
    # SOCs
    idx = lp_len + free_len
    for i in range(0, len(q)):
        z[idx : idx + q[i]] = proj_soc(z[idx : idx + q[i]])
        idx = idx + q[i]
    # SDCs
    for i in range(0, len(s)):
        sz = get_sd_cone_size(s[i])
        z[idx : idx + sz] = proj_sdp(z[idx : idx + sz], s[i])
        idx = idx + sz
    # Exp primal
    for i in range(0, c["ep"]):
        z[idx : idx + 3] = project_exp_bisection(z[idx : idx + 3])
        idx = idx + 3
    # Exp dual
    for i in range(0, c["ed"]):
        z[idx : idx + 3] = z[idx : idx + 3] + project_exp_bisection(
            -z[idx : idx + 3]
        )
        idx = idx + 3
    # Power
    for i in range(0, len(p)):
        if p[i] >= 0:  # primal
            z[idx : idx + 3] = proj_pow(z[idx : idx + 3], p[i])
        else:  # dual
            z[idx : idx + 3] = z[idx : idx + 3] + proj_pow(
                -z[idx : idx + 3], -p[i]
            )
        idx = idx + 3
    return z


def proj_soc(tt):
    tt = np.copy(tt)
    if len(tt) == 0:
        return
    elif len(tt) == 1:
        return pos(tt)

    v1 = tt[0]
    v2 = tt[1:]
    if np.linalg.norm(v2) <= -v1:
        v2 = np.zeros(len(v2))
        v1 = 0
    elif np.linalg.norm(v2) > abs(v1):
        v2 = 0.5 * (1 + v1 / np.linalg.norm(v2)) * v2
        v1 = np.linalg.norm(v2)
    tt[0] = v1
    tt[1:] = v2
    return tt


def proj_sdp(z, n):
    z = np.copy(z)
    if n == 0:
        return
    elif n == 1:
        return pos(z)
    tidx = np.triu_indices(n)
    tidx = (tidx[1], tidx[0])
    didx = np.diag_indices(n)

    a = np.zeros((n, n))
    a[tidx] = z
    a = a + np.transpose(a)
    a[didx] = a[didx] / np.sqrt(2.0)

    w, v = np.linalg.eig(a)  # cols of v are eigenvectors
    w = pos(w)
    a = np.dot(v, np.dot(np.diag(w), np.transpose(v)))
    a[didx] = a[didx] / np.sqrt(2.0)
    z = a[tidx]
    return np.real(z)


def proj_pow(v, a):
    CONE_MAX_ITERS = 20
    CONE_TOL = 1e-8

    if v[0] >= 0 and v[1] >= 0 and (v[0] ** a) * (v[1] ** (1 - a)) >= abs(v[2]):
        return v

    if (
        v[0] <= 0
        and v[1] <= 0
        and ((-v[0] / a) ** a) * ((-v[1] / (1 - a)) ** (1 - a)) >= abs(v[2])
    ):
        return np.zeros(
            3,
        )

    xh = v[0]
    yh = v[1]
    zh = v[2]
    rh = abs(zh)
    r = rh / 2
    for iter in range(0, CONE_MAX_ITERS):
        x = calc_x(r, xh, rh, a)
        y = calc_x(r, yh, rh, 1 - a)

        f = calc_f(x, y, r, a)
        if abs(f) < CONE_TOL:
            break

        dxdr = calcdxdr(x, xh, rh, r, a)
        dydr = calcdxdr(y, yh, rh, r, (1 - a))
        fp = calc_fp(x, y, dxdr, dydr, a)

        r = min(max(r - f / fp, 0), rh)

    z = np.sign(zh) * r
    v[0] = x
    v[1] = y
    v[2] = z
    return v


def calc_x(r, xh, rh, a):
    return max(0.5 * (xh + np.sqrt(xh * xh + 4 * a * (rh - r) * r)), 1e-12)


def calcdxdr(x, xh, rh, r, a):
    return a * (rh - 2 * r) / (2 * x - xh)


def calc_f(x, y, r, a):
    return (x**a) * (y ** (1 - a)) - r


def calc_fp(x, y, dxdr, dydr, a):
    return (x**a) * (y ** (1 - a)) * (a * dxdr / x + (1 - a) * dydr / y) - 1


def project_exp_bisection(v):
    v = np.copy(v)
    r = v[0]
    s = v[1]
    t = v[2]
    # v in cl(Kexp)
    if (s > 0 and t > 0 and r <= s * np.log(t / s)) or (
        r <= 0 and s == 0 and t >= 0
    ):
        return v
    # -v in Kexp^*
    if (-r < 0 and r * np.exp(s / r) <= -np.exp(1) * t) or (
        -r == 0 and -s >= 0 and -t >= 0
    ):
        return np.zeros(
            3,
        )
    # special case with analytical solution
    if r < 0 and s < 0:
        v[1] = 0
        v[2] = max(v[2], 0)
        return v

    x = np.copy(v)
    ub, lb = get_rho_ub(v)
    for iter in range(0, 100):
        rho = (ub + lb) / 2
        g, x = calc_grad(v, rho, x)
        if g > 0:
            lb = rho
        else:
            ub = rho
        if ub - lb < 1e-6:
            break
    return x


def get_rho_ub(v):
    lb = 0
    rho = 2 ** (-3)
    g, z = calc_grad(v, rho, v)
    while g > 0:
        lb = rho
        rho = rho * 2
        g, z = calc_grad(v, rho, z)
    ub = rho
    return ub, lb


def calc_grad(v, rho, warm_start):
    x = solve_with_rho(v, rho, warm_start[1])
    if x[1] == 0:
        g = x[0]
    else:
        g = x[0] + x[1] * np.log(x[1] / x[2])
    return g, x


def solve_with_rho(v, rho, w):
    x = np.zeros(3)
    x[2] = newton_exp_onz(rho, v[1], v[2], w)
    x[1] = (1 / rho) * (x[2] - v[2]) * x[2]
    x[0] = v[0] - rho
    return x


def newton_exp_onz(rho, y_hat, z_hat, w):
    t = max(max(w - z_hat, -z_hat), 1e-6)
    for iter in range(0, 100):
        f = (1 / rho**2) * t * (t + z_hat) - y_hat / rho + np.log(t / rho) + 1
        fp = (1 / rho**2) * (2 * t + z_hat) + 1 / t

        t = t - f / fp
        if t <= -z_hat:
            t = -z_hat
            break
        elif t <= 0:
            t = 0
            break
        elif abs(f) < 1e-6:
            break
    return t + z_hat



# nost test suite copied initially from ECOS project
from __future__ import print_function
import platform

import numpy as np
import scs
import scipy.sparse as sp

c = np.array([-1.0])
b = np.array([1.0, -0.0])
A = sp.csc_matrix([1.0, -1.0]).T.tocsc()
P = sp.csc_matrix([1.2345]).tocsc()
data = {"A": A, "P": P, "b": b, "c": c}
cone = {"q": [], "l": 2}

sol = scs.solve(data, cone, use_indirect=False)
print(sol)
sol = scs.solve(data, cone, use_indirect=True)
print(sol)



from __future__ import print_function
import platform


def import_error(msg):
    print()
    print("## IMPORT ERROR:" + msg)
    print()


try:
    import pytest
except ImportError:
    import_error("Please install pytest to run tests.")
    raise

try:
    import scs
except ImportError:
    import_error("You must install the scs module before running tests.")
    raise

try:
    import numpy as np
    from numpy.testing import assert_almost_equal, assert_array_less
except ImportError:
    import_error("Please install numpy.")
    raise

try:
    import scipy.sparse as sp
except ImportError:
    import_error("Please install scipy.")
    raise

# global data structures for problem
c = np.array([-1.0])
b = np.array([1.0, 0.0])
A = sp.csc_matrix([1.0, -1.0]).T.tocsc()
data = {"A": A, "b": b, "c": c}
cone = {"l": 2}


@pytest.mark.parametrize("use_indirect", [False, True])
def test_backwards_compatibility(use_indirect):
    # max x
    # s.t 0 <= x <= 1
    sol = scs.solve(data, cone, use_indirect=use_indirect, verbose=False)
    assert_almost_equal(sol["x"][0], 1.0, decimal=2)


@pytest.mark.parametrize("use_indirect", [False, True])
def test_update(use_indirect):
    # max x
    # s.t 0 <= x <= 1
    solver = scs.SCS(data, cone, use_indirect=use_indirect, verbose=False)
    sol = solver.solve()
    assert_almost_equal(sol["x"][0], 1.0, decimal=2)

    # min x
    # s.t 0 <= x <= 1
    c_new = np.array([1.0])
    solver.update(c=c_new)
    sol = solver.solve()
    assert_almost_equal(sol["x"][0], 0.0, decimal=2)

    # max x
    # s.t -1 <= x <= 1
    b_new = np.array([1.0, 1.0])
    solver.update(b=b_new)
    sol = solver.solve()
    assert_almost_equal(sol["x"][0], -1.0, decimal=2)


@pytest.mark.parametrize("use_indirect", [False, True])
def test_warm_start(use_indirect):
    # max x
    # s.t 0 <= x <= 1
    solver = scs.SCS(data, cone, use_indirect=use_indirect, verbose=False)
    sol = solver.solve()
    assert_almost_equal(sol["x"][0], 1.0, decimal=2)

    sol = solver.solve()
    assert_almost_equal(sol["x"][0], 1.0, decimal=2)
    assert_array_less(sol["info"]["iter"], 10)

    sol = solver.solve(x=None)
    sol = solver.solve(x=np.array([7.0]), y=None, s=None)
    sol = solver.solve(
        x=np.array([7.0]), y=np.array([1.0, 2.0]), s=np.array([3.0, 4.0])
    )

    with pytest.raises(ValueError):
        sol = solver.solve(x=np.array([1.0, 2.0]))



#!/usr/bin/env python
from warnings import warn
from scipy import sparse
import _scs_direct

__version__ = _scs_direct.version()
__sizeof_int__ = _scs_direct.sizeof_int()
__sizeof_float__ = _scs_direct.sizeof_float()

_USE_INDIRECT_DEFAULT = False


# SCS return integers correspond to one of these flags:
# (copied from scs/include/glbopts.h)
INFEASIBLE_INACCURATE = -7  # SCS best guess infeasible
UNBOUNDED_INACCURATE = -6  # SCS best guess unbounded
SIGINT = -5  # interrupted by sig int
FAILED = -4  # SCS failed
INDETERMINATE = -3  # indeterminate (norm too small)
INFEASIBLE = -2  # primal infeasible, dual unbounded
UNBOUNDED = -1  # primal unbounded, dual infeasible
UNFINISHED = 0  # never returned, used as placeholder
SOLVED = 1  # problem solved to desired accuracy
SOLVED_INACCURATE = 2  # SCS best guess solved


# Choose which SCS to import based on settings.
def _select_scs_module(stgs):
    if stgs.pop("gpu", False):  # False by default
        if not stgs.pop("use_indirect", _USE_INDIRECT_DEFAULT):
            raise NotImplementedError(
                "GPU direct solver not yet available, pass `use_indirect=True`."
            )
        import _scs_gpu

        return _scs_gpu

    if stgs.pop("mkl", False):  # False by default
        if stgs.pop("use_indirect", False):
            raise NotImplementedError(
                "MKL indirect solver not yet available, pass `use_indirect=False`."
            )
        import _scs_mkl

        return _scs_mkl

    if stgs.pop("use_indirect", _USE_INDIRECT_DEFAULT):
        import _scs_indirect

        return _scs_indirect

    return _scs_direct


class SCS(object):
    def __init__(self, data, cone, **settings):
        """Initialize the SCS solver.

        @param data     Dictionary containing keys `P`, `A`, `b`, `c`.
        @param cone     Dictionary containing cone information.
        @param settings Settings as kwargs, see docs.

        """
        self._settings = settings
        if not data or not cone:
            raise ValueError("Missing data or cone information")

        if "b" not in data or "c" not in data:
            raise ValueError("Missing one of b, c from data dictionary")
        if "A" not in data:
            raise ValueError("Missing A from data dictionary")

        A = data["A"]
        b = data["b"]
        c = data["c"]

        if A is None or b is None or c is None:
            raise ValueError("Incomplete data specification")

        if not sparse.issparse(A):
            raise TypeError("A is required to be a sparse matrix")
        if not sparse.isspmatrix_csc(A):
            warn(
                "Converting A to a CSC (compressed sparse column) matrix;"
                " may take a while."
            )
            A = A.tocsc()

        if sparse.issparse(b):
            b = b.todense()

        if sparse.issparse(c):
            c = c.todense()

        m = len(b)
        n = len(c)

        if not A.has_sorted_indices:
            A.sort_indices()
        Adata, Aindices, Acolptr = A.data, A.indices, A.indptr
        if A.shape != (m, n):
            raise ValueError("A shape not compatible with b,c")

        Pdata, Pindices, Pcolptr = None, None, None
        if "P" in data:
            P = data["P"]
            if P is not None:
                if not sparse.issparse(P):
                    raise TypeError("P is required to be a sparse matrix")
                if P.shape != (n, n):
                    raise ValueError("P shape not compatible with A,b,c")
                if not sparse.isspmatrix_csc(P):
                    warn(
                        "Converting P to a CSC (compressed sparse column) "
                        "matrix; may take a while."
                    )
                    P = P.tocsc()
                # extract upper triangular component only
                if sparse.tril(P, -1).data.size > 0:
                    P = sparse.triu(P, format="csc")
                if not P.has_sorted_indices:
                    P.sort_indices()
                Pdata, Pindices, Pcolptr = P.data, P.indices, P.indptr

        # Which scs are we using (scs_direct, scs_indirect, ...)
        _scs = _select_scs_module(self._settings)

        # Initialize solver
        self._solver = _scs.SCS(
            (m, n),
            Adata,
            Aindices,
            Acolptr,
            Pdata,
            Pindices,
            Pcolptr,
            b,
            c,
            cone,
            **self._settings
        )

    def solve(self, warm_start=True, x=None, y=None, s=None):
        """Solve the optimization problem.

        @param warm_start   Whether to warm-start. By default the solution of
                            the previous problem is used as the warm-start. The
                            warm-start can be overriden to another value by
                            passing `x`, `y`, `s` args.
        @param x            Primal warm-start override.
        @param y            Dual warm-start override.
        @param s            Slack warm-start override.

        @return dictionary with solution with keys:
             'x' - primal solution
             's' - primal slack solution
             'y' - dual solution
             'info' - information dictionary (see docs)
        """
        return self._solver.solve(warm_start, x, y, s)

    def update(self, b=None, c=None):
        """Update the `b` vector, `c` vector, or both, before another solve.

        After a solve we can reuse the SCS workspace in another solve if the
        only problem data that has changed are the `b` and `c` vectors.

        @param  b   New `b` vector.
        @param  c	New `c` vector.

        """
        self._solver.update(b, c)


# Backwards compatible helper function that simply calls the main API.
def solve(data, cone, **settings):
    solver = SCS(data, cone, **settings)

    # Hack out the warm start data from old API
    x = y = s = None
    if "x" in data:
        x = data["x"]
    if "y" in data:
        y = data["y"]
    if "s" in data:
        s = data["s"]

    return solver.solve(warm_start=True, x=x, y=y, s=s)



from __future__ import print_function
import platform


def import_error(msg):
    print()
    print("## IMPORT ERROR:" + msg)
    print()


try:
    import pytest
except ImportError:
    import_error("Please install pytest to run tests.")
    raise

try:
    import scs
except ImportError:
    import_error("You must install the scs module before running tests.")
    raise

try:
    import numpy as np
    from numpy.testing import assert_almost_equal
except ImportError:
    import_error("Please install numpy.")
    raise

try:
    import scipy.sparse as sp
except ImportError:
    import_error("Please install scipy.")
    raise

# global data structures for problem
c = np.array([-1.0])
b = np.array([1.0, 0.0])
A = sp.csc_matrix([1.0, -1.0]).T.tocsc()
data = {"A": A, "b": b, "c": c}

FAIL = "failure"  # scs code for failure


@pytest.mark.parametrize(
    "cone,use_indirect,expected",
    [
        ({"q": [], "l": 2}, False, 1),
        ({"q": [], "l": 2}, True, 1),
        ({"q": [2], "l": 0}, False, 0.5),
        ({"q": [2], "l": 0}, True, 0.5),
    ],
)
def test_problems(cone, use_indirect, expected):
    solver = scs.SCS(data, cone=cone, use_indirect=use_indirect, verbose=False)
    sol = solver.solve()
    assert_almost_equal(sol["x"][0], expected, decimal=2)


if platform.python_version_tuple() < ("3", "0", "0"):

    @pytest.mark.parametrize(
        "cone,use_indirect,expected",
        [
            ({"q": [], "l": long(2)}, False, 1),
            ({"q": [], "l": long(2)}, True, 1),
            ({"q": [long(2)], "l": 0}, False, 0.5),
            ({"q": [long(2)], "l": 0}, True, 0.5),
        ],
    )
    def test_problems_with_longs(cone, use_indirect, expected):
        sol = scs.solve(
            data, cone=cone, use_indirect=use_indirect, verbose=False
        )
        assert_almost_equal(sol["x"][0], expected, decimal=2)


def test_failures():
    with pytest.raises(TypeError):
        scs.solve()

    with pytest.raises(ValueError):
        scs.solve(data, {"q": [4], "l": -2})

    # disable this until win64 types figured out
    # with pytest.raises(ValueError)
    # scs.solve(data, {'q': [], 'l': 2}, max_iters=-1)

    # python 2.6 and before just cast float to int
    if platform.python_version_tuple() >= ("2", "7", "0"):
        with pytest.raises(ValueError):
            scs.solve(data, {"q": [], "l": 2}, max_iters=1.1)

    with pytest.raises(ValueError):
        sol = scs.solve(data, {"q": [1], "l": 0}, verbose=False)
